# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# image.py - image widgets
# -----------------------------------------------------------------------------
# This file is imported by the backend process in the clutter
# mainloop. Importing and using clutter is thread-safe.
#
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011-2012 Dirk Meyer
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'CairoTexture', 'ImageTexture' ]

import os

import cairo
from gi.repository import Clutter as clutter

import widget

class CairoTexture(widget.Widget):
    """
    Cairo based Texture widget.
    """

    width = height = 0

    def create(self):
        """
        Create the clutter object
        """
        if not self.width or not self.height:
            self.width = self.height = 1
        self.obj = clutter.Actor()
        self.content = clutter.Canvas()
        self.content.set_size(self.width, self.height)
        self.obj.set_content(self.content);
        self.content.connect('draw', self._draw_canvas)
        self.obj.show()

    def draw(self, cr):
        """
        Render the cairo context
        """
        pass

    def _draw_cairotexture(self, texture, cr):
        """
        Callback from clutter to redraw the canvas
        """
        cr.set_operator (cairo.OPERATOR_CLEAR)
        cr.paint()
        cr.set_operator(cairo.OPERATOR_OVER)
        self.draw(cr)
        return True

    def _draw_canvas(self, texture, cr, width, height):
        """
        Callback from clutter to redraw the canvas
        """
        cr.set_operator (cairo.OPERATOR_CLEAR)
        cr.paint()
        cr.set_operator(cairo.OPERATOR_OVER)
        self.draw(cr)
        return True

    def update(self, modified):
        """
        Render the widget
        """
        super(CairoTexture, self).update(modified)
        if ('width' in modified or 'height' in modified) and self.height and self.width:
            self.content.set_size(self.width, self.height)
        for attribute in 'opacity', 'scale_x', 'scale_y', 'anchor_point':
            if attribute in modified:
                modified.pop(attribute)
        if modified:
            self.content.invalidate()


class ImageTexture(widget.Widget):
    """
    Image widget.
    """

    def create(self):
        """
        Create the clutter object
        """
        self.obj = clutter.Texture.new()
        self.obj.show()

    def update(self, modified):
        """
        Render the widget
        """
        super(ImageTexture, self).update(modified)
        if 'sync_data' in modified and self.sync_data:
            filename, delete = self.sync_data
            # TODO: the prepare function should always load the cogl
            # texture in the main thread to avoid blocking the clutter
            # thread. Loading the image here using kaa.imlib2 and
            # putting the raw data on the texture is actually slower
            # sometimes.
            if self.load_async:
                self.obj.set_load_async(True)
            self.obj.set_from_file(filename)
            if delete:
                os.unlink(filename)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# stage - Backend stage class
# -----------------------------------------------------------------------------
# This file is imported by the backend process in the clutter
# mainloop. Importing and using clutter is thread-safe.
#
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011-2012 Dirk Meyer
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

from gi.repository import Clutter as clutter

class Stage(object):

    def create(self):
        """
        Create the clutter stage object
        """
        self.obj = clutter.Stage.get_default()
        self.obj.hide_cursor()
        self.obj.connect('key-press-event', self.handle_key)
        self.keysyms = {}
        for name in dir(clutter):
            if name.startswith('KEY_'):
                if not getattr(clutter, name) in self.keysyms:
                    self.keysyms[getattr(clutter, name)] = []
                self.keysyms[getattr(clutter, name)].append(name[4:])

    def set_active(self, state):
        """
        Set the stage active or inactive
        """
        # set_opacity does not work on the state
        for child in self.obj.get_children():
            child.save_easing_state()
            child.set_easing_duration(200)
            factor = 2.0 if state else 1/2.0
            child.set_opacity(child.get_opacity() * factor)
            child.restore_easing_state()

    def handle_key(self, stage, event):
        """
        Translate clutter keycode to name and emit signal in main loop. This
        function is a callback from clutter.
        """
        for key in self.keysyms.get(event.keyval, []):
            self.server.send_event('key-press', key)

    def ensure_redraw(self):
        self.obj.ensure_redraw()

    def init(self, size):
        """
        Set the size and the base group
        """
        self.obj.set_size(*size)
        self.obj.set_color(clutter.Color.new(0, 0, 0, 0xff))
        self.obj.show()



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# setup.py - Setup script for kaa.candy
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2008-2011 Dirk Meyer, Jason Tackaberry
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

# python imports
import sys
import os
import sys
import stat
try:
    # kaa base imports
    from kaa.distribution.core import Extension, setup
    import kaa.utils
except ImportError:
    print 'kaa.base not installed'
    sys.exit(1)

print 'checking for gobject-introspection ...',
try:
    import gi
    import gi.repository
except ImportError:
    print 'not found'
    sys.exit(1)
print 'ok'

# version checks
def check_version(name, version, major_version=0, minor_version=0):
    print 'checking for gir %s %s ...' % (name, version),
    try:
        gi.require_version(name, version)
    except ValueError, e:
        print 'not found'
        print e
        sys.exit(1)
    exec('from gi.repository import %s as module' % name)
    if hasattr(module, 'MAJOR_VERSION'):
        if major_version and module.MAJOR_VERSION != major_version:
            print 'failed'
            print 'wrong major version: %s.%s' % (module.MAJOR_VERSION, module.MINOR_VERSION)
            sys.exit(1)
        if module.MINOR_VERSION < minor_version:
            print 'failed'
            print 'wrong minor version: %s.%s' % (module.MAJOR_VERSION, module.MINOR_VERSION)
            sys.exit(1)
    else:
        if major_version and int(module._version.split('.')[0]) != major_version:
            print 'failed'
            print 'wrong major version: %s' % module._version
            sys.exit(1)
    print 'ok'

check_version('GObject', '2.0')
check_version('Gst', '1.0')
check_version('Clutter', '1.0', 1, 12)
check_version('ClutterGst', '2.0')

if len(sys.argv) == 2 and sys.argv[1] == 'clean':
    for file in ('build', 'dist', 'src/version.py', 'MANIFEST',
                 'src/backend/gen_libcandy.c'):
        if os.path.isdir(file):
            print 'removing %s' % file
            os.system('rm -rf %s' % file)
        if os.path.isfile(file):
            print 'removing %s' % file
            os.unlink(file)
    sys.exit(0)

# now trigger the python magic
setup(
    module = 'candy',
    version = '0.0.2',
    license = 'GPL',
    summary = 'Fourth generation Canvas System using Clutter as backend',
    namespace_packages = ['kaa']
)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# grid.py - grid widget
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Grid', 'SelectionGrid' ]

# kaa imports
import kaa

# kaa.candy imports
from .. import is_template
from group import AbstractGroup

class Grid(AbstractGroup):
    """
    Grid holding several widgets based on the given items. The grid
    supports scrolling.
    @note: see C{test/flickr.py} for an example
    """
    candyxml_name = 'grid'

    HORIZONTAL, VERTICAL = range(2)

    __items = None

    def __init__(self, pos, size, cell_size, cell_item, items, template,
                 orientation, xpadding=None, ypadding=None, context=None):
        """
        Simple grid widget to show the items based on the template.

        @param pos: (x,y) position of the widget or None
        @param size: (width,height) geometry of the widget.
        @param cell_size: (width,height) of each cell
        @param cell_item: string how the cell item should be added to the context
        @param items: list of objects or object name in the context
        @param template: child template for each cell
        @param orientation: how to arange the grid: Grid.HORIZONTAL or Grid.VERTICAL
        @param xpadding: x value of space between two items. If set to None
            the padding will be calculated based on cell size and widget size
        @param ypadding: y value of space between two items. If set to None
            the padding will be calculated based on cell size and widget size
        @param context: the context the widget is created in
        """
        super(Grid, self).__init__(pos, size, context=context)
        # store arguments for later public use
        self.cell_size = cell_size
        # store arguments for later private use
        self.__orientation = orientation
        if isinstance(items, (str, unicode)):
            # items is a string, get it from the context
            items = self.context.get(items)
        self.items = items
        self.cell_item = cell_item
        self.template = template
        self.item_padding = xpadding, ypadding

    def create_grid(self):
        """
        Setup the grid. After this function has has been called no modifications
        to the grid are possible.

        @todo: make it possible to change the layout during runtime
        """
        # do some calculations
        padding_x, padding_y = self.item_padding
        if padding_x is None:
            # no padding is given. Get the number fitting items
            # and devide the remaining space as paddingand border
            self.num_items_x = int(self.width / self.cell_size[0])
            padding_x = self.width / self.num_items_x - self.cell_size[0]
            self.item_width = self.cell_size[0] + padding_x
        else:
            self.item_width = self.cell_size[0] + padding_x
            # now that we know the sizes check how much items fit
            self.num_items_x = int(self.width / self.item_width)
        if padding_y is None:
            # no padding is given. Get the number fitting items
            # and devide the remaining space as padding and border
            self.num_items_y = int(self.height / self.cell_size[1])
            padding_y = self.height / self.num_items_y - self.cell_size[1]
            self.item_height = self.cell_size[1] + padding_y
        else:
            self.item_height = self.cell_size[1] + padding_y
            # now that we know the sizes check how much items fit
            self.num_items_y = int(self.height / self.item_height)
        self.item_padding = padding_x, padding_y
        # we now center the grid by default
        x0 = (self.width - self.num_items_x * self.item_width + padding_x) / 2
        y0 = (self.height - self.num_items_y * self.item_height + padding_y) / 2
        self.clip = (x0 - padding_x, y0 - padding_y), \
            (self.num_items_x * self.item_width + padding_x, self.num_items_y * self.item_height + padding_y)
        self.location = (0, 0)
        # list of rendered items
        self.item_widgets = {}
        # group of items
        self.item_group = AbstractGroup((x0, y0))
        self.add(self.item_group)
        self.create_grid = None

    def create_item(self, item_num, pos_x, pos_y):
        """
        Render one child
        """
        if item_num < 0 or item_num >= len(self.items):
            self.item_widgets[(pos_x, pos_y)] = None
            return
        # calculate the size where the child should be
        child_x = pos_x * self.item_width
        child_y = pos_y * self.item_height
        context = self.context.copy()
        context[self.cell_item] = self.items[item_num]
        child = self.template(context=context)
        child.x = child_x
        child.y = child_y
        child.width, child.height = self.cell_size
        self.item_group.add(child)
        self.item_widgets[(pos_x, pos_y)] = child
        return child

    def clear(self):
        """
        Clear the grid
        """
        self.item_group.clear()
        self.item_widgets = {}
        self.queue_rendering()

    def sync_prepare(self):
        """
        Prepare widget for the next sync with the backend
        """
        if self.create_grid:
            self.create_grid()
        if not super(Grid, self).sync_prepare():
            return False
        if self.__orientation == Grid.VERTICAL:
            max_x, max_y = self.location
            for y in range(0, max_y + self.num_items_y):
                for x in range(0, max_x + self.num_items_x):
                    item_num = x + y * self.num_items_x
                    if not (x, y) in self.item_widgets:
                        self.create_item(item_num, x, y)
        if self.__orientation == Grid.HORIZONTAL:
            max_x, max_y = self.location
            for x in range(0, max_x + self.num_items_x):
                for y in range(0, max_y + self.num_items_y):
                    item_num = x * self.num_items_y + y
                    if not (x, y) in self.item_widgets:
                        self.create_item(item_num, x, y)
        return super(Grid, self).sync_prepare()

    def sync_context(self):
        """
        Adjust to a new context. Note: the context of the children is
        not synced. If required the grid must be cleared first. This
        may be fixed in future versions of this widget.
        """
        self.items = self.__items_provided

    @property
    def items(self):
        """
        Get list of items
        """
        return self.__items

    @items.setter
    def items(self, items):
        """
        Set list of items
        """
        self.__items_provided = items
        if isinstance(items, (str, unicode)):
            # items is a string, get it from the context
            items = self.context.get(items)
        if self.__items != items:
            if self.__items:
                # we already had a valid list of items. Set the new
                # list of items before clearing in case an inherting
                # class already needs it on clear
                self.__items = items
                self.clear()
            self.__items = items

    @kaa.synchronized()
    def scroll_by(self, (x, y), secs, force=False):
        """
        Scroll by rows and cols cells

        @param x, y: rows and cols to scroll
        @param secs: runtime of the animation
        """
        # This function will force grid creation
        if self.create_grid:
            self.create_grid()
        while not force:
            # check if it possible to go there
            if self.__orientation == Grid.HORIZONTAL:
                num = (self.location[0] + x) * self.num_items_y + \
                      (self.location[1] + y)
            if self.__orientation == Grid.VERTICAL:
                num = (self.location[1] + y) * self.num_items_x + \
                      (self.location[0] + x)
            if num >= 0 and num < len(self.items):
                # there is an item in the upper left corner
                break
            # remove one cell in scroll, start with x and use y if
            # there are no rows to scroll anymore
            if x:
                x -= (x / abs(x))
            else:
                y -= (y / abs(y))
        self.scroll_to((self.location[0] + x, self.location[1] + y), secs)

    @kaa.synchronized()
    def scroll_to(self, (x, y), secs):
        """
        Scroll to row / cell position

        @param x, y: end row and col
        @param secs: runtime of the animation
        """
        if self.create_grid:
            self.create_grid()
        self.location = (x, y)
        pos_x = -x * self.item_width + self.clip[0][0] + self.item_padding[0]
        pos_y = -y * self.item_height + self.clip[0][1] + self.item_padding[1]
        self.item_group.animate('EASE_OUT_CUBIC', secs, x=pos_x, y=pos_y)
        self.queue_rendering()

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the candyxml element for parameter to create the widget. Example::
          <grid width='100' height='100' cell-width='30' cell-height='30'
              cell-item='item' items='listing'>
              <image filename='$item.filename'/>
          </grid>
        There is only one child element allowed, if more is needed you need
        to add a container as child with the real children in it.
        """
        subelement = element[0]
        orientation = Grid.HORIZONTAL
        if element.orientation and element.orientation.lower() == 'vertical':
            orientation = Grid.VERTICAL
        if element.cell_width:
            element.cell_width = int(element.cell_width)
        if element.cell_height:
            element.cell_height = int(element.cell_height)
        return super(Grid, cls).candyxml_parse(element).update(
            template=subelement.xmlcreate(), items=element.items,
            cell_size=(element.cell_width, element.cell_height), cell_item=element.cell_item,
            orientation=orientation)



class SelectionGrid(Grid):
    """
    Grid with selection widget.
    @note: see C{test/beacon.py} for an example
    """

    candyxml_style = 'selection'

    def __init__(self, pos, size, cell_size, cell_item, items, template,
                 selection, orientation, xpadding=None, ypadding=None, context=None):
        """
        Simple grid widget to show the items based on the template.

        @param pos: (x,y) position of the widget or None
        @param size: (width,height) geometry of the widget.
        @param cell_size: (width,height) of each cell
        @param cell_item: string how the cell item should be added to the context
        @param items: list of objects or object name in the context
        @param template: child template for each cell
        @param selection: widget for the selection
        @param orientation: how to arange the grid: Grid.HORIZONTAL or Grid.VERTICAL
        @param xpadding: x value of space between two items. If set to None
            the padding will be calculated based on cell size and widget size
        @param ypadding: y value of space between two items. If set to None
            the padding will be calculated based on cell size and widget size
        @param context: the context the widget is created in

        """
        super(SelectionGrid, self).__init__(pos, size, cell_size, cell_item, items,
            template, orientation, xpadding, ypadding, context)
        if is_template(selection):
            selection = selection()
        self.selection = selection

    def clear(self):
        """
        Clear the grid
        """
        super(SelectionGrid, self).clear()
        self.item_group.add(self.selection)

    def select(self, (x, y), secs):
        """
        Select a cell.

        @param x, y: cell position to select
        @param secs: runtime of the animation
        """
        if self.create_grid:
            self.create_grid()
        pos_x = x * self.item_width + self.selection.grid_adjust_x
        pos_y = y * self.item_height + self.selection.grid_adjust_y
        self.selection.animate('EASE_OUT_CUBIC', secs, x=pos_x, y=pos_y)
        self.queue_rendering()

    def create_grid(self):
        """
        Setup the grid. After this function has has been called no modifications
        to the grid are possible.

        @todo: make it possible to change the layout during runtime
        """
        super(SelectionGrid, self).create_grid()
        self.item_group.add(self.selection)
        self.selection.grid_adjust_x = self.selection.x = \
            (self.item_width - self.item_padding[0] - self.selection.width) / 2
        self.selection.grid_adjust_y = self.selection.y = \
            (self.item_height - self.item_padding[1] - self.selection.height) / 2

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the candyxml element for parameter to create the widget. Example::
          <grid width='100' height='100' cell-width='30' cell-height='30'
              cell-item='item' items='listing'>
              <image filename='$item.filename'/>
              <selection>
                  <rectangle/>
              </selection/>
          </grid>
        There are only the two children element allowed, if more is
        needed you need to add a container as child with the real
        children in it.
        """
        selection = None
        for child in element:
            if child.node == 'selection':
                selection = child[0].xmlcreate()
                element.remove(child)
        return super(SelectionGrid, cls).candyxml_parse(element).update(
            selection=selection)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# group.py - group widget containing other widgets
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'AbstractGroup', 'Group', 'ConditionGroup' ]

# python imports
import logging

# kaa imports
import kaa

# kaa.candy imports
from widget import Widget
from .. import is_template

# get logging object
log = logging.getLogger('kaa.candy')

class AbstractGroup(Widget):
    """
    A simple group of widgets. This class has no XML parser function
    and therefore, may be a better base class for widgets holding
    children.
    """
    candy_backend = 'candy.Group'
    attributes = [ 'clip' ]

    clip = None

    def __init__(self, pos=None, size=None, context=None):
        super(AbstractGroup, self).__init__(pos, size, context)
        self.children = []

    def __sync__(self, tasks):
        """
        Internal function to add the changes to the list of tasks for
        the backend.
        """
        super(AbstractGroup, self).__sync__(tasks)
        for child in self.children:
            child.__sync__(tasks)
        return True

    def queue_rendering(self):
        """
        Queue sync for layout changes
        """
        super(AbstractGroup, self).queue_rendering()
        # Note: changing one widget will mark all anchestors dirty to
        # force the __sync__ method to call the widget. On the other
        # side, marking an AbstractGroup dirty will also mark all
        # variable sized children. Therefore, changing one widget may
        # result in a complete re-checking all widgets on sync. This
        # is a stupid and unneccessary task, but keeping all
        # dependencies in account is very complex and maybe not worth
        # it. So we mark everything dirty, no big deal. The sync
        # method will only sync what really changed and by this we
        # have the overhead in the main app and not the backend.
        for child in self.children:
            if child.variable_size and not child._candy_dirty:
                child.queue_rendering()

    def sync_context(self):
        """
        Adjust to a new context
        """
        context = self.context
        for child in self.children[:]:
            if child.freeze_context:
                continue
            if not child.supports_context(context):
                self.replace(child, child.__template__(context=context))
            else:
                child.context = context

    def sync_layout(self, size):
        """
        Sync layout changes and calculate intrinsic size based on the
        parent's size.
        """
        super(AbstractGroup, self).sync_layout(size)
        children_width = children_height = 0
        for child in self.children:
            intrinsic_size = child.sync_layout(self.size)
            if child.reference_x != 'parent':
                # ignore this child and calculate this later when we
                # know more about the children
                pass
            elif child.xalign == Widget.ALIGN_SHRINK:
                # auto shrink, use intrinsic size
                children_width = max(children_width, child.x + child.intrinsic_size[0])
            else:
                # use defined width
                children_width = max(children_width, child.x + child.width)
            if child.reference_y != 'parent':
                # ignore this child and calculate this later when we
                # know more about the children
                pass
            elif child.yalign == Widget.ALIGN_SHRINK:
                # auto shrink, use intrinsic size
                children_height = max(children_height, child.y + child.intrinsic_size[1])
            else:
                # use defined height
                children_height = max(children_height, child.y + child.height)
        self.intrinsic_size = children_width, children_height
        # now use that calculated size to set the geometry for the
        # children with their size based on the siblings.
        for child in self.children:
            if child.reference_x == 'parent' and child.reference_y == 'parent':
                continue
            width = children_width if child.reference_x != 'parent' else self.size[0]
            height = children_height if child.reference_y != 'parent' else self.size[1]
            child.sync_layout((width, height))

    def sync_prepare(self):
        """
        Prepare widget for the next sync with the backend
        """
        if not super(AbstractGroup, self).sync_prepare():
            return False
        for child in self.children:
            child.sync_prepare()
        return True

    def get_widget(self, name):
        """
        Get child element with the given name. For group children this
        function will search recursive.

        @param name: name of the child
        @returns: widget or None
        """
        for child in self.children:
            if child.name == name:
                return child
            if isinstance(child, AbstractGroup):
                result = child.get_widget(name)
                if result is not None:
                    return result
        return None

    def get_widget_by_id(self, wid):
        """
        Get child element with the given id. For group children this
        function will search recursive.

        @param wid: id of the child
        @returns: widget or None
        """
        for child in self.children:
            if child._candy_id == wid:
                return child
            if isinstance(child, AbstractGroup):
                result = child.get_widget_by_id(wid)
                if result is not None:
                    return result
        return None

    def add(self, *widgets):
        """
        Add widgets to the group.
        """
        for widget in widgets:
            widget.parent = self

    def remove(self, *widgets):
        """
        Remove widgets from the group.
        """
        for widget in widgets:
            widget.parent = None

    @kaa.coroutine()
    def replace(self, child, replacement):
        """
        Replace the given child with the replacement. Emit the child's
        replace event.
        """
        replacement.parent = self
        replacement._candy_stack = child._candy_id
        if child in Widget._candy_sync_reparent:
            # The old child already is in replacement mode, put the
            # new one where the old one is. FIXME: there are some
            # strange reasons where this does not work as expected. I
            # have no idea why and it is hard to create.
            replacement._candy_stack = child._candy_stack
            Widget._candy_sync_reparent.remove(replacement)
            Widget._candy_sync_reparent.insert(Widget._candy_sync_reparent.index(child)+1, replacement)
            self.children.remove(replacement)
            self.children.insert(self.children.index(child), replacement)
        child.freeze_context = True
        try:
            replacing = child.emit('replace', child, replacement)
            if isinstance(replacing, kaa.InProgress):
                yield replacing
        except Exception, e:
            log.exception('replacing error')
        child.freeze_context = False
        child.parent = None

    def clear(self):
        """
        Clear the group by removing all children
        """
        for widget in self.children[:]:
            widget.parent = None



class Group(AbstractGroup):
    """
    Group widget with XML support
    """

    candyxml_name = 'group'

    def __init__(self, pos=None, size=None, widgets=[], context=None):
        super(Group, self).__init__(pos, size, context)
        for widget in widgets:
            if is_template(widget):
                widget = widget(context)
            self.add(widget)

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the candyxml element for parameter to create the widget. Example::
            <group x='10' y='0' width='200' height=100'>
                <child_widget1/>
                <child_widget2/>
            </group>
        """
        widgets = []
        parameter = super(Group, cls).candyxml_parse(element).update(widgets=widgets)
        for child in element:
            try:
                widget = child.xmlcreate()
            except Exception, e:
                log.error('unable to parse %s: %s', child.node, e)
                continue
            if not widget:
                log.error('unable to parse %s', child.node)
                continue
            widgets.append(widget)
        return parameter


class ConditionGroup(AbstractGroup):
    """
    Group widget switch/case with XML support
    """

    candyxml_name = 'group'
    candyxml_style = 'condition'

    def __init__(self, pos=None, size=None, conditions=[], context=None):
        super(ConditionGroup, self).__init__(pos, size, context)
        for pos, (condition, value, widgets) in enumerate(conditions):
            if isinstance(condition, (str, unicode)):
                condition = self.context.get(condition)
            if value:
                condition = not cmp(str(condition).lower(), str(value).lower())
            if condition:
                for widget in widgets:
                    if is_template(widget):
                        widget = widget(context)
                    self.add(widget)
                break
        else:
            pos = -1
        self.__condition = pos, conditions

    def supports_context(self, context):
        """
        Check if the widget is capable of the given context based on its
        dependencies.
        """
        if not super(ConditionGroup, self).supports_context(context):
            return False
        old, conditions = self.__condition
        for pos, (condition, value, widgets) in enumerate(conditions):
            if isinstance(condition, (str, unicode)):
                condition = context.get(condition)
            if value:
                condition = not cmp(str(condition).lower(), str(value).lower())
            if condition:
                return pos == old
        # No condition matches. Only valid if we had the "problem"
        # before when the widget was created
        return old == -1

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the candyxml element for parameter to create the widget.
        """
        conditions = []
        parameter = super(ConditionGroup, cls).candyxml_parse(element).update(conditions=conditions)
        for c in element:
            widgets = []
            for child in c:
                try:
                    widget = child.xmlcreate()
                except Exception, e:
                    log.error('unable to parse %s: %s', child.node, e)
                    continue
                if not widget:
                    log.error('unable to parse %s', child.node)
                    continue
                widgets.append(widget)
            conditions.append((c.condition or True, c.value, widgets))
        return parameter



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# label.py - label widget
# -----------------------------------------------------------------------------
# This file is imported by the backend process in the clutter
# mainloop. Importing and using clutter is thread-safe.
#
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011-2013 Dirk Meyer
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Label' ]

# cairo imports
import cairo

# kaa.candy imports
import image
import core

class Label(image.CairoTexture):

    def draw(self, cr):
        """
        Render the cairo context
        """
        fade = self.font.get_width(self.text) > self.width
        # draw new text string
        if self.color:
            cr.set_source_rgba(*self.color.to_cairo())
        cr.select_font_face(self.font.name, cairo.FONT_SLANT_NORMAL)
        cr.set_font_size(self.font.size)
        if fade and self.color:
            s = cairo.LinearGradient(0, 0, self.width, 0)
            c = self.color.to_cairo()
            s.add_color_stop_rgba(0, *c)
            # 50 pixel fading
            s.add_color_stop_rgba(1 - (50.0 / self.width), *c)
            s.add_color_stop_rgba(1, c[0], c[1], c[2], 0)
            cr.set_source(s)
        cr.move_to(0, cr.font_extents()[0])
        cr.show_text(self.text)
        return True



import kaa
import kaa.candy
import feedparser

# we use the following xml (file) as gui. There are two main widgets:
# a label called "wait" and a group called "flickr" with a label and
# a grid in it. The grid needs a template for each cell which is again a
# group with a label and an image. The flickr group depends on
# two variables in the context: title for the label and items for the
# grid. The image also has a reflection modifier.
# Note: this reflection is the reason why the scrolling is not smooth
# because we render in software. Remove that line for a better result.
xml = '''
<candyxml geometry="800x600">
    <label name="wait" y="100" font="Vera:24" color="0xcccccc">
        <properties xalign="center"/>
        Loading feed, please wait
    </label>
    <group name="flickr" x="10" y="10" width="780" height="580">
        <label font="Vera:24" color="0xcccccc">
            <properties xalign="center"/>
            $title
        </label>
        <grid y="50" height="530" cell-width="160" cell-height="140"
            items="items" cell-item="item" orientation="vertical">
            <properties name="items"/>
            <group>
                <image url="$item.thumbnail" width="160" height="100">
                    <properties xalign="center" yalign="center" keep-aspect="true"/>
                </image>
                <label y="110" font="Vera:10" color="0xcccccc">
                    <properties xalign="center"/>
                    $item.title
                </label>
            </group>
        </grid>
    </group>
</candyxml>
'''

# create a stage window and parse the xml file
stage = kaa.candy.Stage((800,600), 'flickr')
attr, candy = stage.candyxml(xml)

# add the wait widget to the stage. Since it is only a template it is
# safe to do this in the mainloop.
label = candy.label.wait()
stage.add(label)

# now load a flickr RSS feed and create the context
class Image(object):
    def __init__(self, title, thumbnail):
        self.title = title
        self.thumbnail = thumbnail

@kaa.threaded()
def load_feed(tag):
    # feed = feedparser.parse('feed.xml')
    feed = feedparser.parse('http://api.flickr.com/services/feeds/photos_public.gne?' +
                            'tags=%s&lang=en-us&format=atom' % tag)

    items = []
    num = 0
    for item in feed.entries:
        tmp = item.content[0]['value'][item.content[0]['value'].find('img src="')+9:]
        url = tmp[:tmp.find('"')]
        items.append(Image('%s' % num, url))
        num += 1
    return feed, items


@kaa.coroutine()
def main():
    feed, items = yield load_feed('beach')

    # this is the context for the flickr widget
    context = dict(title=feed.feed.title, items=items)

    # remove the wait label (it is safe to remove something from the stage in the
    # mainloop) and add the flickr group based on the context.
    stage.remove(label)
    group = candy.group.flickr(context=context)
    stage.add(group)

    print 'take a look'
    yield kaa.delay(1)
    print 'scroll down'
    grid = group.get_widget('items')
    grid.scroll_by((0, 2), 1)
    yield kaa.delay(2)
    print 'scroll right'
    grid.scroll_by((2, 0), 4)
    yield kaa.delay(2)
    print 'scroll up very fast, more than possible'
    grid.scroll_by((0,-15), 0.5)
    yield kaa.delay(3)
    print 'and down again'
    grid.scroll_by((1,1), 2)
    yield kaa.delay(0.7)
    print 'and left again while the animation is still running'
    grid.scroll_by((-1, 0), 0.8)
    yield kaa.delay(2)
    print 'go home (0,0)'
    grid.scroll_to((0, 0), 0.8)
    yield kaa.delay(2)
    print 'load more'
    # create new context and replace it
    feed, items = yield load_feed('sunset')
    # set new context
    group.context = kaa.candy.Context(title=feed.feed.title, items=items)

main()

# run the kaa mainloop, it takes some time to load all the images.
kaa.main.run()
